import asyncio
import json
from typing import Dict, Optional, Union

import kserve
import kserve.errors as errors
from common.logging import setup_logging
from common.serialization import (
    DeepNumpyEncoder,
    truss_msgpack_deserialize,
    truss_msgpack_serialize,
)
from fastapi import Depends, FastAPI, Request
from fastapi.responses import ORJSONResponse
from fastapi.routing import APIRoute as FastAPIRoute
from kserve.handlers import DataPlane, ModelRepositoryExtension, V1Endpoints
from model_wrapper import ModelWrapper
from starlette.responses import Response


async def parse_body(request: Request) -> bytes:
    """
    Used by FastAPI to read body in an asynchronous manner
    """
    return await request.body()


class BasetenEndpoints(V1Endpoints):
    def __init__(
        self,
        dataplane: DataPlane,
        model_repository_extension: Optional[ModelRepositoryExtension] = None,
    ):
        super().__init__(dataplane, model_repository_extension)

    @staticmethod
    def check_healthy(model: ModelWrapper):
        if model.load_failed():
            raise errors.InferenceError("Model load failed")

        if not model.ready:
            raise errors.ModelNotReady(model.name)

    async def model_ready(self, model_name: str) -> Dict[str, Union[str, bool]]:
        model: ModelWrapper = self.dataplane.get_model_from_registry(model_name)

        self.check_healthy(model)

        return {}

    def predict(
        self, model_name: str, request: Request, body_raw: bytes = Depends(parse_body)
    ) -> Response:
        """
        This method is called by FastAPI, which introspects that it's not async, and schedules it on a thread
        """
        model: ModelWrapper = self.dataplane.get_model_from_registry(model_name)

        self.check_healthy(model)

        body: Dict
        if self.is_binary(request):
            body = truss_msgpack_deserialize(body_raw)
        else:
            body = json.loads(body_raw)

        # calls kserve.model.Model.__call__, which runs validate, preprocess, predict, and postprocess
        response: Dict = asyncio.run(model(body, headers=dict(request.headers.items())))

        response_headers = {}
        if self.is_binary(request):
            response_headers["Content-Type"] = "application/octet-stream"
            return Response(
                content=truss_msgpack_serialize(response), headers=response_headers
            )
        else:
            response_headers["Content-Type"] = "application/json"
            return Response(
                content=json.dumps(response, cls=DeepNumpyEncoder),
                headers=response_headers,
            )

    @staticmethod
    def is_binary(request: Request):
        return (
            "Content-Type" in request.headers
            and request.headers["Content-Type"] == "application/octet-stream"
        )


class TrussServer(kserve.ModelServer):

    _endpoints: BasetenEndpoints
    _model: ModelWrapper
    _config: Dict

    def __init__(self, http_port: int, config: Dict):
        super().__init__(
            http_port=http_port,
            enable_grpc=False,
            workers=1,
            enable_docs_url=False,
            enable_latency_logging=False,
        )

        self._config = config
        self._endpoints = BasetenEndpoints(
            self.dataplane, self.model_repository_extension
        )

    def start_model(self) -> None:
        """
        Overloaded version of super().start to use instance model in TrussServer
        """
        super().start([])

    def on_startup(self):
        """
        This method will be started inside the main process, so here is where we want to setup our logging and model
        """
        setup_logging()

        self._model = ModelWrapper(self._config)
        self.register_model(self._model)

        self._model.start_load()

    def create_application(self):
        return FastAPI(
            title="Baseten Inference Server",
            docs_url=None,
            redoc_url=None,
            default_response_class=ORJSONResponse,
            on_startup=[self.on_startup],
            routes=[
                # liveness endpoint
                FastAPIRoute(r"/", self.dataplane.live),
                # readiness endpoint
                FastAPIRoute(
                    r"/v1/models/{model_name}", self._endpoints.model_ready, tags=["V1"]
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}:predict",
                    self._endpoints.predict,
                    methods=["POST"],
                    tags=["V1"],
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}:predict_binary",
                    self._endpoints.predict,
                    methods=["POST"],
                    tags=["V1"],
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}:explain",
                    self._endpoints.explain,
                    methods=["POST"],
                    tags=["V1"],
                ),
            ],
            exception_handlers={
                errors.InferenceError: errors.inference_error_handler,
                errors.ModelNotFound: errors.model_not_found_handler,
                errors.ModelNotReady: errors.model_not_ready_handler,
                NotImplementedError: errors.not_implemented_error_handler,
                Exception: errors.generic_exception_handler,
            },
        )



import logging
import sys

from pythonjsonlogger import jsonlogger

LEVEL: int = logging.INFO

JSON_LOG_HANDLER = logging.StreamHandler(stream=sys.stderr)
JSON_LOG_HANDLER.set_name("json_logger_handler")
JSON_LOG_HANDLER.setLevel(LEVEL)
JSON_LOG_HANDLER.setFormatter(
    jsonlogger.JsonFormatter("%(asctime)s %(levelname)s %(message)s")
)


class HealthCheckFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        # for any health check endpoints, lets skip logging
        return (
            record.getMessage().find("GET / ") == -1
            and record.getMessage().find("GET /v1/models/model ") == -1
        )


def setup_logging() -> None:
    loggers = [logging.getLogger()] + [
        logging.getLogger(name) for name in logging.root.manager.loggerDict
    ]

    for logger in loggers:
        logger.setLevel(LEVEL)
        logger.propagate = False

        setup = False

        # let's not thrash the handlers unnecessarily
        for handler in logger.handlers:
            if handler.name == JSON_LOG_HANDLER.name:
                setup = True

        if not setup:
            logger.handlers.clear()
            logger.addHandler(JSON_LOG_HANDLER)

        # some special handling for request logging
        if logger.name == "uvicorn.access":
            logger.addFilter(HealthCheckFilter())



from pathlib import Path
from tempfile import TemporaryDirectory

from truss.contexts.image_builder.serving_image_builder import (
    ServingImageBuilderContext,
)
from truss.truss_handle import TrussHandle

BASE_DIR = Path(__file__).parent


def test_serving_image_dockerfile_from_user_base_image(custom_model_truss_dir):
    th = TrussHandle(custom_model_truss_dir)
    th.set_base_image("baseten/truss-server-base:3.9-v0.4.3")
    builder_context = ServingImageBuilderContext
    image_builder = builder_context.run(th.spec.truss_dir)
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        image_builder.prepare_image_build_dir(tmp_path)
        with open(tmp_path / "Dockerfile", "r") as f:
            gen_docker_lines = f.readlines()
        with open(
            f"{BASE_DIR}/../../../test_data/context_builder_image_test/server.Dockerfile",
            "r",
        ) as f:
            server_docker_lines = f.readlines()

        def filter_empty_lines(lines):
            return list(filter(lambda x: x and x != "\n", lines))

        gen_docker_lines = filter_empty_lines(gen_docker_lines)
        server_docker_lines = filter_empty_lines(server_docker_lines)
        assert gen_docker_lines == server_docker_lines



import os

import nbformat
import pytest
from nbconvert.preprocessors import ExecutePreprocessor


@pytest.mark.parametrize("notebook", ["happy.ipynb"])
def test_notebook_exec(notebook):
    """
    Test for Jupyter notebooks to establish some base exercising.  Future tests can add new notebooks similar to
    happy.ipynb in the test_data directory, then add that file to the above parameterization
    """
    with open(
        f"{os.path.dirname(os.path.realpath(__file__))}/../test_data/{notebook}"
    ) as f:
        nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600, kernel_name="python3")
        ep.preprocess(nb)



def model_supports_predict_proba(model: object) -> bool:
    if not hasattr(model, "predict_proba"):
        return False
    if hasattr(
        model, "_check_proba"
    ):  # noqa eg Support Vector Machines *can* predict proba if they made certain choices while training
        try:
            model._check_proba()
            return True
        except AttributeError:
            return False
    return True



import os
from collections.abc import Mapping
from pathlib import Path
from typing import Dict


class SecretsResolver:
    SECRETS_MOUNT_DIR = "/secrets"
    SECRET_ENV_VAR_PREFIX = "TRUSS_SECRET_"

    @staticmethod
    def get_secrets(config: Dict):
        return Secrets(config.get("secrets", {}))

    @staticmethod
    def _resolve_secret(secret_name: str, default_value: str):
        secret_value = default_value
        secret_env_var_name = SecretsResolver.SECRET_ENV_VAR_PREFIX + secret_name
        if secret_env_var_name in os.environ:
            secret_value = os.environ[secret_env_var_name]
        secret_path = SecretsResolver._secrets_mount_dir_path() / secret_name
        if secret_path.exists() and secret_path.is_file():
            with secret_path.open() as secret_file:
                secret_value = secret_file.read()
        return secret_value

    @staticmethod
    def _secrets_mount_dir_path():
        return Path(SecretsResolver.SECRETS_MOUNT_DIR)


class Secrets(Mapping):
    def __init__(self, base_secrets: Dict[str, str]):
        self._base_secrets = base_secrets

    def __getitem__(self, key: str) -> str:
        return SecretsResolver._resolve_secret(key, self._base_secrets[key])

    def __iter__(self):
        raise NotImplementedError(
            "Secrets are meant for lookup and can't be iterated on"
        )

    def __len__(self):
        return len(self._base_secrets)



from typing import Any

import torch
from torch import package


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        config = kwargs["config"]
        self._model_metadata = config["model_metadata"]
        self._model = None
        self._model_dtype = None
        self._device = torch.device(
            "cuda:0"
            if torch.cuda.is_available() and config["resources"]["use_gpu"]
            else "cpu"
        )

    def load(self):
        imp = package.PackageImporter(
            self._data_dir / "model" / self._model_metadata["torch_package_file"]
        )
        package_name = self._model_metadata["torch_model_package_name"]
        model_pickle_filename = self._model_metadata["torch_model_pickle_filename"]
        self._model = imp.load_pickle(package_name, model_pickle_filename)
        self._model_dtype = list(self._model.parameters())[0].dtype

    def predict(self, model_input: Any) -> Any:
        model_output = {}
        with torch.no_grad():
            inputs = torch.tensor(
                model_input, dtype=self._model_dtype, device=self._device
            )
            model_output["predictions"] = self._model(inputs).tolist()
            return model_output



from pathlib import Path
from typing import Dict, Set

from truss.constants import PYTORCH_REQ_MODULE_NAMES
from truss.model_framework import ModelFramework
from truss.types import ModelFrameworkType

TORCH_PACKAGE_FILE = "model_package.pt"
TORCH_MODEL_PICKLE_FILENAME = "model.pkl"
TORCH_MODEL_PACKAGE_NAME = "torch_model"


class PyTorch(ModelFramework):
    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.PYTORCH

    def required_python_depedencies(self) -> Set[str]:
        return PYTORCH_REQ_MODULE_NAMES

    def serialize_model_to_directory(self, model, target_directory: Path):
        from torch import package

        try:
            _torch_package(model, target_directory / TORCH_PACKAGE_FILE, [])
        except package.package_exporter.PackagingError as pkg_err:
            # Make sure previous, potentially partially written, package is gone
            (target_directory / TORCH_PACKAGE_FILE).unlink()
            modules_to_extern = _broken_modules(pkg_err)
            _torch_package(
                model, target_directory / TORCH_PACKAGE_FILE, modules_to_extern
            )

    def model_metadata(self, model) -> Dict[str, str]:
        return {
            "torch_package_file": TORCH_PACKAGE_FILE,
            "torch_model_pickle_filename": TORCH_MODEL_PICKLE_FILENAME,
            "torch_model_package_name": TORCH_MODEL_PACKAGE_NAME,
            "model_binary_dir": "model",
        }

    def supports_model_class(self, model_class) -> bool:
        try:
            import torch

            return issubclass(model_class, torch.nn.Module)
        except ModuleNotFoundError:
            return False


def _torch_package(model, path: Path, extern_modules) -> None:
    from torch import package

    with package.PackageExporter(path) as exp:
        exp.intern(f"{model.__class__.__module__}.**")
        for extern_module in extern_modules:
            exp.extern(f"{extern_module}.**")
        exp.save_pickle(TORCH_MODEL_PACKAGE_NAME, TORCH_MODEL_PICKLE_FILENAME, model)


def _broken_modules(pkg_error):
    """Extract the broken modules from the torch package error.

    We would extern these modules.
    Args:
        pkg_error (package.package_exporter.PackagingError): Error to extract broken modules from.
    """
    return [
        module_name
        for module_name, attrs in pkg_error.dependency_graph.nodes.items()
        if attrs.get("error") is not None
    ]



import logging
import os
import signal
import subprocess
from typing import List, Optional

from helpers.context_managers import current_directory


class InferenceServerProcessController:

    _inference_server_process: Optional[subprocess.Popen] = None
    _inference_server_port: int
    _inference_server_home: str
    _app_logger: logging.Logger
    _inference_server_process_args: List[str]

    def __init__(
        self,
        inference_server_home: str,
        inference_server_process_args: List[str],
        inference_server_port: int,
        app_logger: logging.Logger,
    ) -> None:
        self._inference_server_home = inference_server_home
        self._inference_server_process_args = inference_server_process_args
        self._inference_server_port = inference_server_port
        self._inference_server_started = False
        self._inference_server_ever_started = False
        self._app_logger = app_logger

    def start(self):
        with current_directory(self._inference_server_home):
            inf_env = os.environ.copy()
            inf_env["INFERENCE_SERVER_PORT"] = str(self._inference_server_port)
            self._inference_server_process = subprocess.Popen(
                self._inference_server_process_args,
                env=inf_env,
            )

            self._inference_server_started = True
            self._inference_server_ever_started = True

    def stop(self):
        if self._inference_server_process is not None:
            name = " ".join(self._inference_server_process_args)

            for line in os.popen("ps ax | grep '" + name + "' | grep -v grep"):
                pid = line.split()[0]
                os.kill(int(pid), signal.SIGKILL)

        self._inference_server_started = False

    def inference_server_started(self) -> bool:
        return self._inference_server_started

    def inference_server_ever_started(self) -> bool:
        return self._inference_server_ever_started

    def is_inference_server_running(self) -> bool:
        # Explicitly check if inference server process is up, this is a bit expensive.
        if not self._inference_server_started:
            return False

        if self._inference_server_process is None:
            return False

        return self._inference_server_process.poll() is None

    def check_and_recover_inference_server(self):
        if self.inference_server_started() and not self.is_inference_server_running():
            self._app_logger.warning(
                "Inference server seems to have crashed, restarting"
            )
            self.start()



import pickle
from typing import Any


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        self._config = kwargs["config"]
        self._pipeline = None

    def load(self):
        with open(self._data_dir / "pipeline.cpick", "rb") as f:
            self._pipeline = pickle.load(f)

    def predict(self, model_input: Any) -> Any:
        return self._pipeline(model_input)



import logging
import os
import sys
from pathlib import Path
from typing import Any, Callable, List, Optional

import click
import cloudpickle
import yaml
from truss.constants import CONFIG_FILE, TEMPLATES_DIR, TRUSS
from truss.docker import kill_containers
from truss.environment_inference.requirements_inference import infer_deps
from truss.errors import FrameworkNotSupportedError
from truss.model_frameworks import MODEL_FRAMEWORKS_BY_TYPE, model_framework_from_model
from truss.model_inference import infer_python_version, map_to_supported_python_version
from truss.notebook import is_notebook_or_ipython
from truss.truss_config import DEFAULT_EXAMPLES_FILENAME, TrussConfig
from truss.truss_handle import TrussHandle
from truss.types import ModelFrameworkType
from truss.util.gpu import get_gpu_memory
from truss.util.path import build_truss_target_directory, copy_file_path, copy_tree_path

logger: logging.Logger = logging.getLogger(__name__)

if is_notebook_or_ipython():
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.StreamHandler(sys.stdout))


def populate_target_directory(
    config: TrussConfig,
    target_directory_path: Optional[str] = None,
    template: str = "custom",
) -> Path:
    target_directory_path_typed = None
    if target_directory_path is None:
        target_directory_path_typed = build_truss_target_directory(template)
    else:
        target_directory_path_typed = Path(target_directory_path)
        target_directory_path_typed.mkdir(parents=True, exist_ok=True)

    # Create data dir
    (target_directory_path_typed / config.data_dir).mkdir()

    # Create bundled packages dir
    (target_directory_path_typed / config.bundled_packages_dir).mkdir()

    # Create model module dir
    model_dir = target_directory_path_typed / config.model_module_dir
    template_path = TEMPLATES_DIR / template
    copy_tree_path(template_path / "model", model_dir)

    examples_path = template_path / DEFAULT_EXAMPLES_FILENAME
    if examples_path.exists():
        copy_file_path(
            examples_path, target_directory_path_typed / DEFAULT_EXAMPLES_FILENAME
        )

    # Write config
    with (target_directory_path_typed / CONFIG_FILE).open("w") as config_file:
        yaml.dump(config.to_dict(), config_file)

    return target_directory_path_typed


def create_from_model(
    model: Any,
    target_directory: Optional[str] = None,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
) -> TrussHandle:
    """Create a Truss with the given model. A Truss is a build context designed to
    be built as a container locally or uploaded into a model serving environment.

    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a keras, sklearn, or pytorch model
            object)
        target_directory (str, optional): The local directory target for the Truss. Otherwise a temporary directory
            will be generated
        data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves to
            files for the root directory or a directory path.
        requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
            container environment.
        bundled_packages (List[str], optional): Additional local packages that are required by the model.
    Returns:
        TrussHandle: A handle to the generated Truss that provides easy access to content inside.
    """
    model_framework = model_framework_from_model(model)
    if model_framework.typ() == ModelFrameworkType.XGBOOST:
        click.echo(
            click.style(
                """WARNING: Truss uses XGBoost save/load which has a
                different interface during inference than the class
                you used to train this model. You can learn more about
                these differences at
                https://xgboost.readthedocs.io/en/stable/tutorials/saving_model.html
                """,
                fg="yellow",
            )
        )
    if target_directory is None:
        target_directory_path = build_truss_target_directory(
            model_framework.typ().value
        )
    else:
        target_directory_path = Path(target_directory)
    model_framework.to_truss(model, target_directory_path)
    scaf = TrussHandle(target_directory_path)
    _update_truss_props(scaf, data_files, requirements_file, bundled_packages)
    return scaf


def create_from_pipeline(
    pipeline: Callable,
    target_directory: Optional[str] = None,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
):
    """Create a Truss from a function. A Truss is a build context designed to
    be built as a container locally or uploaded into a model serving environment.

    Args:
        pipeline (a callable function): A function that is expected to be called
        when the Truss server /predict is invoked.
        target_directory (str, optional): The local directory target for the Truss. Otherwise a temporary directory
            will be generated
        data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves to
            files for the root directory or a directory path.
        requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
            container environment.
        bundled_packages (List[str], optional): Additional local packages that are required by the model.
    Returns:
        TrussHandle: A handle to the generated Truss that provides easy access to content inside.
    """

    # Create Truss config
    requirements = list(infer_deps(must_include_deps=set(["cloudpickle"])))
    config = TrussConfig(
        model_type="custom",
        model_framework=ModelFrameworkType.CUSTOM,
        python_version=map_to_supported_python_version(infer_python_version()),
        requirements=requirements,
    )

    # Create and populate target directory path
    target_directory_path = populate_target_directory(
        config=config, target_directory_path=target_directory, template="pipeline"
    )

    gpu_memory = get_gpu_memory()
    if gpu_memory and gpu_memory > 10:
        # TODO: Abu: Remove use of click here
        click.echo(
            click.style(
                """WARNING: Truss identified objects in GPU memory. When serializing a
                function via create(), objects in GPU memory must be moved to
                CPU to be serialized correctly.""",
                fg="yellow",
            )
        )

    # Write Cloudpickled function to data directory
    pipeline_binary_path = target_directory_path / config.data_dir / "pipeline.cpick"
    with open(pipeline_binary_path, "wb") as f:
        cloudpickle.dump(pipeline, f)

    scaf = TrussHandle(target_directory_path)
    _update_truss_props(scaf, data_files, requirements_file, bundled_packages)
    return scaf


def create_from_mlflow_uri(
    model_uri: str,
    target_directory: Optional[str] = None,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
):
    """Create a Truss with the given model. A Truss is a build context designed to
    be built as a container locally or uploaded into a model serving environment.

    Args:
        model_uri (str): URI pointing to the MLflow model.
        target_directory (str, optional): The local directory target for the Truss. Otherwise a temporary directory
            will be generated
        data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves to
            files for the root directory or a directory path.
        requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
            container environment.
        bundled_packages (List[str], optional): Additional local packages that are required by the model.
    Returns:
        TrussHandle: A handle to the generated Truss that provides easy access to content inside.
    """
    model_framework = MODEL_FRAMEWORKS_BY_TYPE[ModelFrameworkType.MLFLOW]
    if target_directory is None:
        target_directory_path = build_truss_target_directory(
            model_framework.typ().value
        )
    else:
        target_directory_path = Path(target_directory)
    model_framework.to_truss(model_uri, target_directory_path)
    truss = TrussHandle(target_directory_path)
    _update_truss_props(truss, data_files, requirements_file, bundled_packages)
    return truss


def create_from_model_with_exception_handler(*args):
    # returns None if framework not supported, otherwise the Truss
    try:
        return create_from_model(*args)
    except FrameworkNotSupportedError:
        return None


def init(
    target_directory: str,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
    trainable: bool = False,
) -> TrussHandle:
    """
    Initialize an empty placeholder Truss. A Truss is a build context designed
    to be built as a container locally or uploaded into a baseten serving
    environment. This placeholder structure can be filled to represent ML
    models.

    Args:
        target_directory: Absolute or relative path of the directory to create
                          Truss in. The directory is created if it doesn't exist.
    """
    config = TrussConfig(
        model_type="custom",
        model_framework=ModelFrameworkType.CUSTOM,
        python_version=map_to_supported_python_version(infer_python_version()),
    )

    target_directory_path = populate_target_directory(
        config=config, target_directory_path=target_directory
    )

    if trainable:
        _populate_default_training_code(
            config, target_directory_path=Path(target_directory)
        )

    scaf = TrussHandle(target_directory_path)
    _update_truss_props(scaf, data_files, requirements_file, bundled_packages)
    return scaf


def load(truss_directory: str) -> TrussHandle:
    """Get a handle to a Truss. A Truss is a build context designed to be built
    as a container locally or uploaded into a model serving environment.

    Args:
        truss_directory (str): The local directory of an existing Truss
    Returns:
        TrussHandle
    """
    return TrussHandle(Path(truss_directory))


def from_directory(*args, **kwargs):
    logger.warn(
        "DeprecationWarning: from_directory() is deprecated. Use load() instead."
    )
    return load(*args, **kwargs)


def create(
    model: Any,
    target_directory: Optional[str] = None,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
) -> TrussHandle:
    # Some model objects can are callable (like Keras models)
    # so we first attempt to make Truss via a model object

    model_scaffold = create_from_model_with_exception_handler(
        model, target_directory, data_files, requirements_file, bundled_packages
    )
    if model_scaffold:
        return model_scaffold
    else:
        if callable(model):
            return create_from_pipeline(
                model, target_directory, data_files, requirements_file, bundled_packages
            )

    raise ValueError(
        "Invalid input to make Truss. Truss expects a supported framework or callable function."
    )


def mk_truss(*args, **kwargs):
    logger.warn("DeprecationWarning: mk_truss() is deprecated. Use create() instead.")
    return create(*args, **kwargs)


def cleanup() -> None:
    """
    Cleans up .truss directory.
    """
    build_folder_path = Path(Path.home(), ".truss")
    if build_folder_path.exists():
        for obj in build_folder_path.glob("**/*"):
            if (not obj.name == "config.yaml") and (obj.is_file()):
                os.remove(obj)
    return


def _update_truss_props(
    scaf: TrussHandle,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
) -> None:
    if data_files is not None:
        for data_file in data_files:
            scaf.add_data(data_file)

    if bundled_packages is not None:
        for package in bundled_packages:
            scaf.add_bundled_package(package)

    if requirements_file is not None:
        scaf.update_requirements_from_file(requirements_file)


def _populate_default_training_code(
    config: TrussConfig,
    target_directory_path: Path,
) -> None:
    """Populate default training code in a truss.

    Assumes target directory already exists.
    """
    # TODO(pankaj): Add support customization based on model framework type, for
    # now we don't support this.
    truss_template = "custom"
    template_path = TEMPLATES_DIR / truss_template
    truss_training_module_dir = target_directory_path / config.train.training_module_dir
    copy_tree_path(template_path / "train", truss_training_module_dir)


def kill_all() -> None:
    kill_containers({TRUSS: True})



import inspect
from pathlib import Path
from typing import Dict, Optional

from truss.contexts.local_loader.truss_module_loader import truss_module_loaded
from truss.contexts.local_loader.utils import (
    prepare_secrets,
    signature_accepts_keyword_arg,
)
from truss.contexts.truss_context import TrussContext
from truss.truss_spec import TrussSpec


class LocalTrainer(TrussContext):
    """Allows training a truss locally."""

    @staticmethod
    def run(truss_dir: Path):
        def train(variables: Optional[Dict] = None):
            spec = TrussSpec(truss_dir)
            with truss_module_loaded(
                str(truss_dir),
                spec.train_module_fullname,
                spec.bundled_packages_dir.name,
                [str(path.resolve()) for path in spec.external_package_dirs_paths],
            ) as module:
                train_class = getattr(module, spec.train_class_name)
                train_class_signature = inspect.signature(train_class)
                train_init_params = {}
                config = spec.config
                if signature_accepts_keyword_arg(train_class_signature, "config"):
                    train_init_params["config"] = config.to_dict()
                if signature_accepts_keyword_arg(train_class_signature, "output_dir"):
                    train_init_params["output_dir"] = truss_dir / "data"
                if signature_accepts_keyword_arg(train_class_signature, "secrets"):
                    train_init_params["secrets"] = prepare_secrets(spec)

                # Wire up variables
                if signature_accepts_keyword_arg(train_class_signature, "variables"):
                    runtime_variables = {}
                    runtime_variables.update(config.train.variables)
                    if variables is not None:
                        runtime_variables.update(variables)
                    train_init_params["variables"] = runtime_variables
                trainer = train_class(**train_init_params)

                if hasattr(trainer, "pre_train"):
                    trainer.pre_train()

                trainer.train()

                if hasattr(trainer, "post_train"):
                    trainer.post_train()

        return train



from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from truss.constants import HTTP_PUBLIC_BLOB_BACKEND
from truss.errors import ValidationError
from truss.types import ModelFrameworkType
from truss.util.data_structures import transform_optional
from truss.validation import (
    validate_cpu_spec,
    validate_memory_spec,
    validate_secret_name,
)

DEFAULT_MODEL_FRAMEWORK_TYPE = ModelFrameworkType.CUSTOM
DEFAULT_MODEL_TYPE = "Model"
DEFAULT_MODEL_MODULE_DIR = "model"
DEFAULT_BUNDLED_PACKAGES_DIR = "packages"
DEFAULT_MODEL_CLASS_FILENAME = "model.py"
DEFAULT_MODEL_CLASS_NAME = "Model"
DEFAULT_TRUSS_STRUCTURE_VERSION = "2.0"
DEFAULT_MODEL_INPUT_TYPE = "Any"
DEFAULT_PYTHON_VERSION = "py39"
DEFAULT_DATA_DIRECTORY = "data"
DEFAULT_EXAMPLES_FILENAME = "examples.yaml"
DEFAULT_SPEC_VERSION = "2.0"
DEFAULT_SPEC_VERSION_ON_LOAD = "1.0"

DEFAULT_CPU = "500m"
DEFAULT_MEMORY = "512Mi"
DEFAULT_USE_GPU = False

DEFAULT_TRAINING_CLASS_FILENAME = "train.py"
DEFAULT_TRAINING_CLASS_NAME = "Train"
DEFAULT_TRAINING_MODULE_DIR = "train"

DEFAULT_BLOB_BACKEND = HTTP_PUBLIC_BLOB_BACKEND


class Accelerator(Enum):
    T4 = "T4"
    A10G = "A10G"
    V100 = "V100"
    A100 = "A100"


@dataclass
class AcceleratorSpec:
    accelerator: Optional[Accelerator] = None
    count: int = 0

    def to_str(self) -> Optional[str]:
        if self.accelerator is None or self.count == 0:
            return None
        if self.count > 1:
            return f"{self.accelerator.value}:{self.count}"
        return self.accelerator.value

    @staticmethod
    def from_str(acc_spec: Optional[str]):
        if acc_spec is None:
            return AcceleratorSpec()
        parts = acc_spec.split(":")
        count = 1
        if len(parts) not in [1, 2]:
            raise ValidationError("`accelerator` does not match parsing requirements.")
        if len(parts) == 2:
            count = int(parts[1])
        try:
            acc = Accelerator[parts[0]]
        except KeyError as exc:
            raise ValidationError(f"Accelerator {acc_spec} not supported") from exc
        return AcceleratorSpec(accelerator=acc, count=count)


@dataclass
class Resources:
    cpu: str = DEFAULT_CPU
    memory: str = DEFAULT_MEMORY
    use_gpu: bool = DEFAULT_USE_GPU
    accelerator: AcceleratorSpec = AcceleratorSpec()

    @staticmethod
    def from_dict(d):
        cpu = d.get("cpu", DEFAULT_CPU)
        validate_cpu_spec(cpu)
        memory = d.get("memory", DEFAULT_MEMORY)
        validate_memory_spec(memory)
        accelerator = AcceleratorSpec.from_str((d.get("accelerator", None)))
        use_gpu = d.get("use_gpu", DEFAULT_USE_GPU)
        if accelerator.accelerator is not None:
            use_gpu = True

        return Resources(
            cpu=cpu,
            memory=memory,
            use_gpu=use_gpu,
            accelerator=accelerator,
        )

    def to_dict(self):
        return {
            "cpu": self.cpu,
            "memory": self.memory,
            "use_gpu": self.use_gpu,
            "accelerator": self.accelerator.to_str(),
        }


@dataclass
class Train:
    training_class_filename: str = DEFAULT_TRAINING_CLASS_FILENAME
    training_class_name: str = DEFAULT_TRAINING_CLASS_NAME
    training_module_dir: str = DEFAULT_TRAINING_MODULE_DIR
    variables: Dict = field(default_factory=dict)
    resources: Resources = field(default_factory=Resources)

    @staticmethod
    def from_dict(d):
        return Train(
            training_class_filename=d.get(
                "training_class_filename", DEFAULT_TRAINING_CLASS_FILENAME
            ),
            training_class_name=d.get(
                "training_class_name", DEFAULT_TRAINING_CLASS_NAME
            ),
            training_module_dir=d.get(
                "training_module_dir", DEFAULT_TRAINING_MODULE_DIR
            ),
            variables=d.get("variables", {}),
            resources=Resources.from_dict(d.get("resources", {})),
        )

    def to_dict(self):
        return {
            "training_class_filename": self.training_class_filename,
            "training_class_name": self.training_class_name,
            "training_module_dir": self.training_module_dir,
            "variables": self.variables,
            "resources": self.resources.to_dict(),
        }


@dataclass
class ExternalDataItem:
    """A piece of remote data, to be made available to the Truss at serving time.

    Remote data is downloaded and stored under Truss's data directory. Care should be taken
    to avoid conflicts. This will get precedence if there's overlap.
    """

    # Url to download the data from.
    # Currently only files are allowed.
    url: str

    # This should be path relative to data directory. This is where the remote
    # file will be downloaded.
    local_data_path: str

    # The backend used to download files
    backend: str = DEFAULT_BLOB_BACKEND

    # A name can be given to a data item for readability purposes. It's not used
    # in the download process.
    name: Optional[str] = None

    @staticmethod
    def from_dict(d: Dict[str, str]) -> "ExternalDataItem":
        url = d.get("url")
        if url is None or url == "":
            raise ValueError("URL of an external data item cannot be empty")
        local_data_path = d.get("local_data_path")
        if local_data_path is None or local_data_path == "":
            raise ValueError(
                "The `local_data_path` field of an external data item cannot be empty"
            )

        item = ExternalDataItem(
            url=d["url"],
            local_data_path=d["local_data_path"],
            name=d.get("name"),
            backend=d.get("backend", DEFAULT_BLOB_BACKEND),
        )
        return item

    def to_dict(self):
        d = {
            "url": self.url,
            "local_data_path": self.local_data_path,
            "backend": self.backend,
        }
        if self.name is not None:
            d["name"] = self.name
        return d


@dataclass
class ExternalData:
    """[Experimental] External data is data that is not contained in the Truss folder.

    Typically this will be data stored remotely. This data is guaranteed to be made
    available under the data directory of the truss.
    """

    items: List[ExternalDataItem]

    @staticmethod
    def from_list(items: List[Dict[str, str]]) -> "ExternalData":
        return ExternalData([ExternalDataItem.from_dict(item) for item in items])

    def to_list(self) -> List[Dict[str, str]]:
        return [item.to_dict() for item in self.items]


@dataclass
class TrussConfig:
    model_framework: ModelFrameworkType = DEFAULT_MODEL_FRAMEWORK_TYPE
    model_type: str = DEFAULT_MODEL_TYPE
    model_name: Optional[str] = None

    model_module_dir: str = DEFAULT_MODEL_MODULE_DIR
    model_class_filename: str = DEFAULT_MODEL_CLASS_FILENAME
    model_class_name: str = DEFAULT_MODEL_CLASS_NAME

    data_dir: str = DEFAULT_DATA_DIRECTORY
    external_data: Optional[ExternalData] = None

    # Python types for what the model expects as input
    input_type: str = DEFAULT_MODEL_INPUT_TYPE
    model_metadata: Dict[str, Any] = field(default_factory=dict)
    requirements: List[str] = field(default_factory=list)
    system_packages: List[str] = field(default_factory=list)
    environment_variables: Dict[str, str] = field(default_factory=dict)
    resources: Resources = field(default_factory=Resources)
    python_version: str = DEFAULT_PYTHON_VERSION
    examples_filename: str = DEFAULT_EXAMPLES_FILENAME
    secrets: Dict[str, str] = field(default_factory=dict)
    description: Optional[str] = None
    bundled_packages_dir: str = DEFAULT_BUNDLED_PACKAGES_DIR
    external_package_dirs: List[str] = field(default_factory=list)
    live_reload: bool = False
    # spec_version is a version string
    spec_version: str = DEFAULT_SPEC_VERSION
    train: Train = field(default_factory=Train)
    base_image: Optional[str] = None

    @property
    def canonical_python_version(self) -> str:
        return {
            "py39": "3.9",
            "py38": "3.8",
            "py37": "3.7",
        }[self.python_version]

    @staticmethod
    def from_dict(d):
        config = TrussConfig(
            # Users that are calling `load` on an existing Truss
            # should default to 1.0 whereas users creating a new Truss
            # should default to 2.0.
            spec_version=d.get("spec_version", DEFAULT_SPEC_VERSION_ON_LOAD),
            model_type=d.get("model_type", DEFAULT_MODEL_TYPE),
            model_framework=ModelFrameworkType(
                d.get("model_framework", DEFAULT_MODEL_FRAMEWORK_TYPE.value)
            ),
            model_module_dir=d.get("model_module_dir", DEFAULT_MODEL_MODULE_DIR),
            model_class_filename=d.get(
                "model_class_filename", DEFAULT_MODEL_CLASS_FILENAME
            ),
            model_class_name=d.get("model_class_name", DEFAULT_MODEL_CLASS_NAME),
            data_dir=d.get("data_dir", DEFAULT_DATA_DIRECTORY),
            input_type=d.get("input_type", DEFAULT_MODEL_INPUT_TYPE),
            model_metadata=d.get("model_metadata", {}),
            requirements=d.get("requirements", []),
            system_packages=d.get("system_packages", []),
            environment_variables=d.get("environment_variables", {}),
            resources=Resources.from_dict(d.get("resources", {})),
            python_version=d.get("python_version", DEFAULT_PYTHON_VERSION),
            model_name=d.get("model_name", None),
            examples_filename=d.get("examples_filename", DEFAULT_EXAMPLES_FILENAME),
            secrets=d.get("secrets", {}),
            description=d.get("description", None),
            bundled_packages_dir=d.get(
                "bundled_packages_dir", DEFAULT_BUNDLED_PACKAGES_DIR
            ),
            external_package_dirs=d.get("external_package_dirs", []),
            live_reload=d.get("live_reload", False),
            train=Train.from_dict(d.get("train", {})),
            external_data=transform_optional(
                d.get("external_data"), ExternalData.from_list
            ),
            base_image=d.get("base_image", None),
        )
        config.validate()
        return config

    @staticmethod
    def from_yaml(yaml_path: Path):
        with yaml_path.open() as yaml_file:
            return TrussConfig.from_dict(yaml.safe_load(yaml_file))

    def write_to_yaml_file(self, path: Path):
        with path.open("w") as config_file:
            yaml.dump(self.to_dict(), config_file)

    def to_dict(self):
        d = {
            "model_type": self.model_type,
            "model_framework": self.model_framework.value,
            "model_module_dir": self.model_module_dir,
            "model_class_filename": self.model_class_filename,
            "model_class_name": self.model_class_name,
            "data_dir": self.data_dir,
            "input_type": self.input_type,
            "model_metadata": self.model_metadata,
            "requirements": self.requirements,
            "system_packages": self.system_packages,
            "environment_variables": self.environment_variables,
            "resources": self.resources.to_dict(),
            "python_version": self.python_version,
            "model_name": self.model_name,
            "examples_filename": self.examples_filename,
            "secrets": self.secrets,
            "description": self.description,
            "bundled_packages_dir": self.bundled_packages_dir,
            "external_package_dirs": self.external_package_dirs,
            "live_reload": self.live_reload,
            "spec_version": self.spec_version,
            "train": self.train.to_dict(),
            "base_image": self.base_image,
        }
        if self.external_data is not None:
            d["external_data"] = transform_optional(
                self.external_data, lambda data: data.to_list()
            )
        return d

    def clone(self):
        return TrussConfig.from_dict(self.to_dict())

    def validate(self):
        for secret_name in self.secrets:
            validate_secret_name(secret_name)



from typing import List

import pandas as pd  # noqa
from truss.build import create


def test_infer_deps_through_create_local_import(sklearn_rfc_model, tmp_path):
    dir_path = tmp_path / "truss"
    import requests  # noqa

    tr = create(
        sklearn_rfc_model,
        target_directory=dir_path,
    )
    spec = tr.spec
    _validate_that_package_is_in_requirements(spec.requirements, "requests")


def test_infer_deps_through_create_top_of_the_file_import(sklearn_rfc_model, tmp_path):
    dir_path = tmp_path / "truss"
    tr = create(
        sklearn_rfc_model,
        target_directory=dir_path,
    )
    spec = tr.spec
    _validate_that_package_is_in_requirements(spec.requirements, "pandas")


def _validate_that_package_is_in_requirements(
    requirements_list: List[str],
    package_name: str,
):
    requirement_entries = [
        req for req in requirements_list if req.startswith(package_name)
    ]
    assert len(requirement_entries) == 1



from pathlib import Path
from typing import Optional

from truss.constants import (
    BASE_SERVER_REQUIREMENTS_TXT_FILENAME,
    CONTROL_SERVER_CODE_DIR,
    MODEL_DOCKERFILE_NAME,
    REQUIREMENTS_TXT_FILENAME,
    SERVER_CODE_DIR,
    SERVER_DOCKERFILE_TEMPLATE_NAME,
    SERVER_REQUIREMENTS_TXT_FILENAME,
    SHARED_SERVING_AND_TRAINING_CODE_DIR,
    SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
    SYSTEM_PACKAGES_TXT_FILENAME,
    TEMPLATES_DIR,
)
from truss.contexts.image_builder.image_builder import ImageBuilder
from truss.contexts.image_builder.util import (
    TRUSS_BASE_IMAGE_VERSION_TAG,
    file_is_not_empty,
    to_dotted_python_version,
    truss_base_image_name,
    truss_base_image_tag,
)
from truss.contexts.truss_context import TrussContext
from truss.patch.hash import directory_content_hash
from truss.truss_spec import TrussSpec
from truss.util.jinja import read_template_from_fs
from truss.util.path import (
    build_truss_target_directory,
    copy_tree_or_file,
    copy_tree_path,
)

BUILD_SERVER_DIR_NAME = "server"
BUILD_CONTROL_SERVER_DIR_NAME = "control"


class ServingImageBuilderContext(TrussContext):
    @staticmethod
    def run(truss_dir: Path):
        return ServingImageBuilder(truss_dir)


class ServingImageBuilder(ImageBuilder):
    def __init__(self, truss_dir: Path) -> None:
        self._truss_dir = truss_dir
        self._spec = TrussSpec(truss_dir)

    @property
    def default_tag(self):
        return f"{self._spec.model_framework_name}-model:latest"

    def prepare_image_build_dir(self, build_dir: Optional[Path] = None):
        """
        Prepare a directory for building the docker image from.
        """
        truss_dir = self._truss_dir
        spec = self._spec
        config = spec.config
        model_framework_name = spec.model_framework_name
        if build_dir is None:
            # TODO(pankaj) We probably don't need model framework specific directory.
            build_dir = build_truss_target_directory(model_framework_name)

        def copy_into_build_dir(from_path: Path, path_in_build_dir: str):
            copy_tree_or_file(from_path, build_dir / path_in_build_dir)  # type: ignore[operator]

        # Copy over truss
        copy_tree_path(truss_dir, build_dir)

        # Copy inference server code
        copy_into_build_dir(SERVER_CODE_DIR, BUILD_SERVER_DIR_NAME)
        copy_into_build_dir(
            SHARED_SERVING_AND_TRAINING_CODE_DIR,
            BUILD_SERVER_DIR_NAME + "/" + SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
        )

        # Copy control server code
        if config.live_reload:
            copy_into_build_dir(CONTROL_SERVER_CODE_DIR, BUILD_CONTROL_SERVER_DIR_NAME)

        # Copy base TrussServer requirements if supplied custom base image
        if config.base_image:
            base_truss_server_reqs_filepath = (
                SERVER_CODE_DIR / REQUIREMENTS_TXT_FILENAME
            )
            copy_into_build_dir(
                base_truss_server_reqs_filepath, BASE_SERVER_REQUIREMENTS_TXT_FILENAME
            )

        # Copy model framework specific requirements file
        server_reqs_filepath = (
            TEMPLATES_DIR / model_framework_name / REQUIREMENTS_TXT_FILENAME
        )
        should_install_server_requirements = file_is_not_empty(server_reqs_filepath)
        if should_install_server_requirements:
            copy_into_build_dir(server_reqs_filepath, SERVER_REQUIREMENTS_TXT_FILENAME)

        (build_dir / REQUIREMENTS_TXT_FILENAME).write_text(spec.requirements_txt)
        (build_dir / SYSTEM_PACKAGES_TXT_FILENAME).write_text(spec.system_packages_txt)

        self._render_dockerfile(build_dir, should_install_server_requirements)

    def _render_dockerfile(
        self,
        build_dir: Path,
        should_install_server_requirements: bool,
    ):
        config = self._spec.config
        data_dir = build_dir / config.data_dir
        bundled_packages_dir = build_dir / config.bundled_packages_dir
        dockerfile_template = read_template_from_fs(
            TEMPLATES_DIR, SERVER_DOCKERFILE_TEMPLATE_NAME
        )
        python_version = to_dotted_python_version(config.python_version)
        if config.base_image:
            base_image_name_and_tag = config.base_image
        else:
            base_image_name = truss_base_image_name(job_type="server")
            tag = truss_base_image_tag(
                python_version=python_version,
                use_gpu=config.resources.use_gpu,
                live_reload=config.live_reload,
                version_tag=TRUSS_BASE_IMAGE_VERSION_TAG,
            )
            base_image_name_and_tag = f"{base_image_name}:{tag}"
        should_install_system_requirements = file_is_not_empty(
            build_dir / SYSTEM_PACKAGES_TXT_FILENAME
        )
        should_install_python_requirements = file_is_not_empty(
            build_dir / REQUIREMENTS_TXT_FILENAME
        )
        dockerfile_contents = dockerfile_template.render(
            should_install_server_requirements=should_install_server_requirements,
            base_image_name_and_tag=base_image_name_and_tag,
            should_install_system_requirements=should_install_system_requirements,
            should_install_requirements=should_install_python_requirements,
            config=config,
            python_version=python_version,
            live_reload=config.live_reload,
            data_dir_exists=data_dir.exists(),
            bundled_packages_dir_exists=bundled_packages_dir.exists(),
            truss_hash=directory_content_hash(self._truss_dir),
        )
        docker_file_path = build_dir / MODEL_DOCKERFILE_NAME
        docker_file_path.write_text(dockerfile_contents)



import logging
import re
from pathlib import Path
from typing import Dict

from endpoints import control_app
from flask import Flask
from helpers.errors import PatchApplicatonError
from helpers.inference_server_controller import InferenceServerController
from helpers.inference_server_process_controller import InferenceServerProcessController
from helpers.patch_applier import PatchApplier
from werkzeug.exceptions import HTTPException


def create_app(base_config: Dict):
    app = Flask(__name__)
    # TODO(pankaj): change this back to info once things are stable
    app.logger.setLevel(logging.DEBUG)
    app.config.update(base_config)
    app.config[
        "inference_server_process_controller"
    ] = InferenceServerProcessController(
        app.config["inference_server_home"],
        app.config["inference_server_process_args"],
        app.config["inference_server_port"],
        app_logger=app.logger,
    )
    patch_applier = PatchApplier(
        Path(app.config["inference_server_home"]),
        app.logger,
        app.config.get("pip_path"),
    )
    app.config["inference_server_controller"] = InferenceServerController(
        app.config["inference_server_process_controller"],
        patch_applier,
        app.logger,
        app.config.get("oversee_inference_server", True),
    )
    app.register_blueprint(control_app)

    def handle_error(exc):
        try:
            raise exc
        except HTTPException:
            return exc
        except PatchApplicatonError:
            app.logger.exception(exc)
            error_type = _camel_to_snake_case(type(exc).__name__)
            return {
                "error": {
                    "type": error_type,
                    "msg": str(exc),
                }
            }
        except Exception:
            app.logger.exception(exc)
            return {
                "error": {
                    "type": "unknown",
                    "msg": f"{type(exc)}: {exc}",
                }
            }

    app.register_error_handler(Exception, handle_error)
    return app


def _camel_to_snake_case(camel_cased: str) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "_", camel_cased).lower()



import os
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List

import pytest

# Needed to simulate the set up on the model docker container
sys.path.append(
    str(
        Path(__file__).parent.parent.parent.parent.parent
        / "templates"
        / "control"
        / "control"
    )
)

from truss.templates.control.control.application import create_app  # noqa
from truss.templates.control.control.helpers.types import (  # noqa
    Action,
    ModelCodePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
)


@pytest.fixture
def truss_original_hash():
    return "1234"


@pytest.fixture
def app(truss_container_fs, truss_original_hash):
    with _env_var({"HASH_TRUSS": truss_original_hash}):
        inf_serv_home = truss_container_fs / "app"
        control_app = create_app(
            {
                "inference_server_home": inf_serv_home,
                "inference_server_process_args": ["python", "inference_server.py"],
                "control_server_host": "0.0.0.0",
                "control_server_port": 8081,
                "inference_server_port": 8082,
                "oversee_inference_server": False,
                "pip_path": "pip",
            }
        )
        inference_server_controller = control_app.config["inference_server_controller"]
        try:
            inference_server_controller.start()
            yield control_app
        finally:
            inference_server_controller.stop()


@pytest.fixture()
def client(app):
    return app.test_client()


def test_restart_server(client):
    resp = client.post("/control/stop_inference_server")
    assert resp.status_code == 200
    assert "error" not in resp.json
    assert "msg" in resp.json

    # Try second restart
    resp = client.post("/control/stop_inference_server")
    assert resp.status_code == 200
    assert "error" not in resp.json
    assert "msg" in resp.json


def test_patch_model_code_update_existing(app, client):
    mock_model_file_content = """
class Model:
    def predict(self, request):
        return {'prediction': [1]}
"""
    patch = Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.UPDATE,
            path="model.py",
            content=mock_model_file_content,
        ),
    )
    _verify_apply_patch_success(client, patch)
    with (
        app.config["inference_server_home"] / "model" / "model.py"
    ).open() as model_file:
        new_model_file_content = model_file.read()
    assert new_model_file_content == mock_model_file_content


def test_patch_model_code_update_predict_on_long_load_time(app, client):
    mock_model_file_content = """
class Model:
    def load(self):
        import time
        time.sleep(3)

    def predict(self, request):
        return {'prediction': [1]}
"""
    patch = Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.UPDATE,
            path="model.py",
            content=mock_model_file_content,
        ),
    )
    _verify_apply_patch_success(client, patch)
    resp = client.post("/v1/models/model:predict", json={})
    resp.status_code == 200
    assert resp.json == {"prediction": [1]}


def test_patch_model_code_create_new(app, client):
    empty_content = ""
    patch = Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.UPDATE,
            path="touched",
            content=empty_content,
        ),
    )
    _verify_apply_patch_success(client, patch)
    assert (app.config["inference_server_home"] / "model" / "touched").exists()


def test_patch_model_code_create_in_new_dir(app, client):
    empty_content = ""
    patch = Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.UPDATE,
            path="new_directory/touched",
            content=empty_content,
        ),
    )
    _verify_apply_patch_success(client, patch)
    assert (
        app.config["inference_server_home"] / "model" / "new_directory" / "touched"
    ).exists()


def test_404(client):
    resp = client.post("/control/nonexitant")
    assert resp.status_code == 404


def test_invalid_patch(client):
    patch_request = {
        "hash": "dummy",
        "prev_hash": "invalid",
        "patches": [],
    }
    resp = client.post("/control/patch", json=patch_request)
    assert resp.status_code == 200
    assert "error" in resp.json
    assert resp.json["error"]["type"] == "inadmissible_patch"
    assert "msg" not in resp.json


def test_unsupported_patch(client):
    unsupported_patch = {
        "type": "unsupported",
        "body": {},
    }
    resp = _apply_patches(client, [unsupported_patch])
    assert resp.status_code == 200
    assert "error" in resp.json
    assert resp.json["error"]["type"] == "unsupported_patch"


def test_patch_failed_recoverable(client):
    will_fail_patch = Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=Action.ADD, requirement="not_a_valid_python_requirement"
        ),
    )
    resp = _apply_patches(client, [will_fail_patch.to_dict()])
    assert resp.status_code == 200
    assert "error" in resp.json
    assert resp.json["error"]["type"] == "patch_failed_recoverable"


def test_patch_failed_unrecoverable(client):
    will_pass_patch = Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(action=Action.ADD, requirement="requests"),
    )
    will_fail_patch = Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=Action.ADD, requirement="not_a_valid_python_requirement"
        ),
    )
    resp = _apply_patches(
        client, [will_pass_patch.to_dict(), will_fail_patch.to_dict()]
    )
    assert resp.status_code == 200
    assert "error" in resp.json
    assert resp.json["error"]["type"] == "patch_failed_unrecoverable"


def _verify_apply_patch_success(client, patch: Patch):
    original_hash = client.get("/control/truss_hash").json["result"]
    patch_request = {
        "hash": "dummy",
        "prev_hash": original_hash,
        "patches": [patch.to_dict()],
    }
    resp = client.post("/control/patch", json=patch_request)
    resp = _apply_patches(client, [patch.to_dict()])
    assert resp.status_code == 200
    assert "error" not in resp.json
    assert "msg" in resp.json


def _apply_patches(client, patches: List[dict]):
    original_hash = client.get("/control/truss_hash").json["result"]
    patch_request = {
        "hash": "dummy",
        "prev_hash": original_hash,
        "patches": patches,
    }
    return client.post("/control/patch", json=patch_request)


@contextmanager
def _env_var(kvs: Dict[str, str]):
    orig_env = os.environ.copy()
    try:
        os.environ.update(kvs)
        yield
    finally:
        os.environ.clear()
        os.environ.update(orig_env)



import shutil
import subprocess
from pathlib import Path
from typing import List, Tuple

import requests

BLOB_DOWNLOAD_TIMEOUT_SECS = 600  # 10 minutes
B10CP_PATH = "/app/bin/b10cp"


def download_external_data(data_dir: Path, config: dict):
    for item in _external_data_items(data_dir, config):
        _download(*item)


def _external_data_items(data_dir: Path, config: dict) -> List[Tuple[str, Path]]:
    items: List[Tuple[str, Path]] = []
    if "external_data" not in config:
        return items

    for item in config["external_data"]:
        backend = item.get("backend", "http_public")
        item_url = item["url"]
        local_data_path = item["local_data_path"]
        if backend != "http_public":
            raise ValueError(f"Unknown backend {backend}")
        items.append((item_url, data_dir / local_data_path))
    return items


def _download(URL: str, download_to: Path):
    download_to.parent.mkdir(exist_ok=True, parents=True)

    b10cp_success = _try_b10cp(URL, download_to)
    if b10cp_success:
        return

    # Fallback if b10cp can't handle the url
    _download_using_requests(URL, download_to)


def _download_using_requests(URL: str, download_to: Path):
    # Streaming download to keep memory usage low
    resp = requests.get(
        URL,
        allow_redirects=True,
        stream=True,
        timeout=BLOB_DOWNLOAD_TIMEOUT_SECS,
    )
    resp.raise_for_status()
    with download_to.open("wb") as file:
        shutil.copyfileobj(resp.raw, file)


def _try_b10cp(URL: str, download_to: Path) -> bool:
    if not Path(B10CP_PATH).exists():
        return False

    proc = subprocess.run(
        [
            B10CP_PATH,
            "-source",
            URL,  # Add quotes to work with any special characters.
            "-target",
            str(download_to),
        ],
        check=False,
    )
    return proc.returncode == 0



import subprocess
from pathlib import Path
from typing import Optional

from helpers.errors import UnsupportedPatch
from helpers.types import (
    Action,
    ModelCodePatch,
    Patch,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_config import TrussConfig


class PatchApplier:
    def __init__(
        self,
        inference_server_home: Path,
        app_logger,
        pip_path: Optional[str] = None,  # Only meant for testing
    ) -> None:
        self._inference_server_home = inference_server_home
        self._model_module_dir = (
            self._inference_server_home / self._truss_config.model_module_dir
        )
        self._app_logger = app_logger
        self._pip_path_cached = None
        if pip_path is not None:
            self._pip_path_cached = "pip"

    def apply_patch(self, patch: Patch):
        self._app_logger.debug(f"Applying patch {patch.to_dict()}")
        if isinstance(patch.body, ModelCodePatch):
            model_code_patch: ModelCodePatch = patch.body
            self._apply_model_code_patch(model_code_patch)
        elif isinstance(patch.body, PythonRequirementPatch):
            py_req_patch: PythonRequirementPatch = patch.body
            self._apply_python_requirement_patch(py_req_patch)
        elif isinstance(patch.body, SystemPackagePatch):
            sys_pkg_patch: SystemPackagePatch = patch.body
            self._apply_system_package_patch(sys_pkg_patch)
        else:
            raise UnsupportedPatch(f"Unknown patch type {patch.type}")

    @property
    def _truss_config(self) -> TrussConfig:
        return TrussConfig.from_yaml(self._inference_server_home / "config.yaml")

    @property
    def _pip_path(self) -> str:
        if self._pip_path_cached is None:
            self._pip_path_cached = _identify_pip_path()
        return self._pip_path_cached

    def _apply_model_code_patch(self, model_code_patch: ModelCodePatch):
        self._app_logger.debug(
            f"Applying model code patch {model_code_patch.to_dict()}"
        )
        action = model_code_patch.action
        filepath: Path = self._model_module_dir / model_code_patch.path
        if action in [Action.ADD, Action.UPDATE]:
            filepath.parent.mkdir(parents=True, exist_ok=True)
            self._app_logger.info(f"Updating file {filepath}")
            with filepath.open("w") as file:
                content = model_code_patch.content
                if content is None:
                    raise ValueError(
                        "Invalid patch: content of a model code update patch should not be None."
                    )
                file.write(content)

        elif action == Action.REMOVE:
            if not filepath.exists():
                self._app_logger.warning(
                    f"Could not delete file {filepath}: not found."
                )
            else:
                self._app_logger.info(f"Deleting file {filepath}")
                filepath.unlink()
        else:
            raise ValueError(f"Unknown model code patch action {action}")

    def _apply_python_requirement_patch(
        self, python_requirement_patch: PythonRequirementPatch
    ):
        self._app_logger.debug(
            f"Applying python requirement patch {python_requirement_patch.to_dict()}"
        )
        action = python_requirement_patch.action

        if action == Action.REMOVE:
            subprocess.run(
                [
                    self._pip_path,
                    "uninstall",
                    "-y",
                    python_requirement_patch.requirement,
                ],
                check=True,
            )
        elif action in [Action.ADD, Action.UPDATE]:
            subprocess.run(
                [
                    self._pip_path,
                    "install",
                    python_requirement_patch.requirement,
                ],
                check=True,
            )
        else:
            raise ValueError(f"Unknown python requirement patch action {action}")

    def _apply_system_package_patch(self, system_package_patch: SystemPackagePatch):
        self._app_logger.debug(
            f"Applying system package patch {system_package_patch.to_dict()}"
        )
        action = system_package_patch.action

        if action == Action.REMOVE:
            subprocess.run(
                [
                    "apt",
                    "remove",
                    "-y",
                    system_package_patch.package,
                ],
                check=True,
            )
        elif action in [Action.ADD, Action.UPDATE]:
            subprocess.run(
                [
                    "apt",
                    "update",
                ],
                check=True,
            )
            subprocess.run(
                [
                    "apt",
                    "install",
                    "-y",
                    system_package_patch.package,
                ],
                check=True,
            )
        else:
            raise ValueError(f"Unknown python requirement patch action {action}")


def _identify_pip_path() -> str:
    if Path("/usr/local/bin/pip3").exists():
        return "/usr/local/bin/pip3"

    if Path("/usr/local/bin/pip").exists():
        return "/usr/local/bin/pip"

    raise RuntimeError("Unable to find pip, make sure it's installed.")



import json
import logging
import os
from functools import wraps
from pathlib import Path
from typing import Callable, List, Optional, Union

import click
import truss
import yaml

logging.basicConfig(level=logging.INFO)


def echo_output(f: Callable[..., object]):
    @wraps(f)
    def wrapper(*args, **kwargs):
        click.echo(f(*args, **kwargs))

    return wrapper


def error_handling(f: Callable[..., object]):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except ValueError as e:
            click.echo(e)
            print_help()
        except Exception as e:
            click.echo(e)

    return wrapper


def print_help() -> None:
    ctx = click.get_current_context()
    click.echo(ctx.get_help())


@click.group(name="truss", invoke_without_command=True)
@click.pass_context
@click.option(
    "-v",
    "--version",
    is_flag=True,
    show_default=False,
    default=False,
    help="Show Truss package version.",
)
def cli_group(ctx, version) -> None:
    if not ctx.invoked_subcommand:
        if version:
            click.echo(truss.version())
        else:
            click.echo(ctx.get_help())


@cli_group.command()
@click.argument("target_directory", required=True)
@click.option(
    "-s",
    "--skip-confirm",
    is_flag=True,
    show_default=True,
    default=False,
    help="Skip confirmation prompt.",
)
@click.option(
    "-t",
    "--trainable",
    is_flag=True,
    show_default=True,
    default=False,
    help="Create a trainable truss.",
)
@error_handling
def init(target_directory, skip_confirm, trainable) -> None:
    """Initializes an empty Truss directory.

    TARGET_DIRECTORY: A Truss is created in this directory
    """
    tr_path = Path(target_directory)
    if skip_confirm or click.confirm(f"A Truss will be created at {tr_path}"):
        truss.init(target_directory=target_directory, trainable=trainable)
        click.echo(f"Truss was created in {tr_path}")


@cli_group.command()
@click.argument("build_dir")
@click.argument("target_directory", required=False)
@error_handling
def build_context(build_dir, target_directory: str) -> None:
    """
    Create a docker build context for a Truss.

    BUILD_DIR: Folder where image context is built for Truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.docker_build_setup(build_dir=Path(build_dir))


@cli_group.command()
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@error_handling
@click.option("--tag", help="Docker image tag")
def build_image(target_directory: str, build_dir: Path, tag) -> None:
    """
    Builds the docker image for a Truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    tr.build_serving_docker_image(build_dir=build_dir, tag=tag)


@cli_group.command()
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@click.option("--tag", help="Docker build image tag")
@click.option("--port", type=int, default=8080, help="Local port used to run image")
@click.option(
    "--attach", is_flag=True, default=False, help="Flag for attaching the process"
)
@error_handling
def run_image(target_directory: str, build_dir: Path, tag, port, attach) -> None:
    """
    Runs the docker image for a Truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    urls = tr.get_urls_from_truss()
    if urls:
        click.confirm(
            f"Container already exists at {urls}. Are you sure you want to continue?"
        )
    tr.docker_run(build_dir=build_dir, tag=tag, local_port=port, detach=not attach)


@cli_group.command()
@click.option("--target_directory", required=False, help="Directory of truss")
@click.option(
    "--request",
    type=str,
    required=False,
    help="String formatted as json that represents request",
)
@click.option(
    "--build-dir",
    type=click.Path(exists=True),
    required=False,
    help="Directory where context is built",
)
@click.option("--tag", help="Docker build image tag")
@click.option("--port", type=int, default=8080, help="Local port used to run image")
@click.option(
    "--use-docker",
    is_flag=True,
    default=True,
    help="Flag to run prediction with a docker container",
)
@click.option(
    "--request-file",
    type=click.Path(exists=True),
    help="Path to json file containing the request",
)
@error_handling
@echo_output
def predict(
    target_directory: str,
    request: Union[bytes, str],
    build_dir,
    tag,
    port,
    use_docker,
    request_file,
):
    """
    Invokes the packaged model, either locally or in a Docker container.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    REQUEST: String formatted as json that represents request

    BUILD_DIR: Directory where context is built. If none, a temp directory is created.

    TAG: Docker build image tag

    PORT: Local port used to run image

    USE_DOCKER: Flag to run prediction with a docker container

    REQUEST_FILE: Path to json file containing the request
    """
    if request is not None:
        request_data = json.loads(request)
    elif request_file is not None:
        with open(request_file) as json_file:
            request_data = json.load(json_file)
    else:
        raise ValueError("At least one of request or request-file must be supplied.")

    tr = _get_truss_from_directory(target_directory=target_directory)
    if use_docker:
        return tr.docker_predict(
            request_data, build_dir=build_dir, tag=tag, local_port=port, detach=True
        )
    else:
        return tr.server_predict(request_data)


@cli_group.command()
@click.option("--target_directory", required=False, help="Directory of truss")
@click.option(
    "--build-dir",
    type=click.Path(exists=True),
    required=False,
    help="Directory where context is built",
)
@click.option("--tag", help="Docker build image tag")
@click.option(
    "--var",
    multiple=True,
    help="""Training variables in key=value form where value is string.
    For more complex values use vars_yaml_file""",
)
@click.option(
    "--vars_yaml_file",
    required=False,
    help="Training variables from a yaml file",
)
@click.option(
    "--local",
    is_flag=True,
    default=False,
    help="Flag to run training locally (not on docker)",
)
@error_handling
@echo_output
def train(target_directory: str, build_dir, tag, var: List[str], vars_yaml_file, local):
    """Runs prediction for a Truss in a docker image or locally"""
    tr = _get_truss_from_directory(target_directory=target_directory)
    if vars_yaml_file is not None:
        with Path(vars_yaml_file).open() as vars_file:
            variables = yaml.safe_load(vars_file)
    else:
        variables = _variables_dict_from_option(var)
    if local:
        return tr.local_train(variables=variables)

    return tr.docker_train(build_dir=build_dir, tag=tag, variables=variables)


@cli_group.command()
@click.argument("target_directory", required=False)
@click.option("--name", type=str, required=False, help="Name of example to run")
@click.option(
    "--local", is_flag=True, default=False, help="Flag to run prediction locally"
)
@error_handling
@echo_output
def run_example(target_directory: str, name, local):
    """
    Runs examples specified in the Truss, over docker.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    predict_fn = tr.docker_predict
    if local:
        predict_fn = tr.server_predict

    if name is not None:
        example = tr.example(name)
        click.echo(f"Running example: {name}")
        return predict_fn(example.input)
    else:
        example_outputs = []
        for example in tr.examples():
            click.echo(f"Running example: {example.name}")
            example_outputs.append(predict_fn(example.input))
        return example_outputs


@cli_group.command()
@click.argument("target_directory", required=False)
@error_handling
def get_container_logs(target_directory) -> None:
    """
    Get logs in a container is running for a truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    for log in _get_truss_from_directory(
        target_directory=target_directory
    ).serving_container_logs():
        click.echo(log)


@cli_group.command()
@click.argument("target_directory", required=False)
def kill(target_directory: str) -> None:
    """
    Kills containers related to truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.kill_container()


@cli_group.command()
def kill_all() -> None:
    "Kills all truss containers that are not manually persisted"
    truss.kill_all()


@cli_group.command()
@error_handling
def cleanup() -> None:
    """
    Clean up truss data.

    Truss creates temporary directories for various operations
    such as for building docker images. This command clears
    that data to free up disk space.
    """
    truss.build.cleanup()


def _get_truss_from_directory(target_directory: Optional[str] = None):
    """Gets Truss from directory. If none, use the current directory"""
    if target_directory is None:
        target_directory = os.getcwd()
    return truss.load(target_directory)


def _variables_dict_from_option(variables_list: List[str]) -> dict:
    vars_dict = {}
    for var in variables_list:
        first_equals_pos = var.find("=")
        if first_equals_pos == -1:
            raise ValueError(
                f"Training variable expected in `key=value` from but found `{var}`",
            )
        var_name = var[:first_equals_pos]
        var_value = var[first_equals_pos + 1 :]
        vars_dict[var_name] = var_value
    return vars_dict


if __name__ == "__main__":
    cli_group()



import tempfile
from pathlib import Path

import pytest
import requests
from python_on_whales import docker
from truss.constants import CONFIG_FILE, TRUSS
from truss.contexts.image_builder.serving_image_builder import (
    ServingImageBuilderContext,
)
from truss.contexts.local_loader.load_model_local import LoadModelLocal
from truss.model_frameworks.keras import Keras
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_config import TrussConfig
from truss.truss_handle import wait_for_truss


def test_to_truss(keras_mpg_model):
    with tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        framework = Keras()
        framework.to_truss(keras_mpg_model, truss_dir)

        # Assertions
        config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
        assert config.model_class_filename == "model.py"
        assert config.model_class_name == "Model"
        assert config.model_framework == framework.typ()
        assert config.model_type == "Model"
        assert config.python_version.startswith("py3")
        assert len(config.requirements) != 0

        model_metadata = config.model_metadata
        assert model_metadata["model_binary_dir"] == "model"

        assert (truss_dir / "data" / "model").exists()
        assert (truss_dir / "model" / "model.py").exists()


def test_run_truss(keras_mpg_model):
    with tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        sklearn_framework = Keras()
        sklearn_framework.to_truss(keras_mpg_model, truss_dir)
        model = LoadModelLocal.run(truss_dir)
        result = model.predict([0, 0, 0, 0, 0, 0, 0, 0, 0])
        predictions = result["predictions"]
        assert len(predictions) == 1
        assert len(predictions[0]) == 1


@pytest.mark.integration
def test_run_image(keras_mpg_model):
    with ensure_kill_all(), tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        framework = Keras()
        framework.to_truss(keras_mpg_model, truss_dir)
        tag = f"test-{framework.typ().value}-model:latest"
        image = ServingImageBuilderContext.run(truss_dir).build_image(
            tag=tag, labels={TRUSS: True}
        )
        assert image.repo_tags == [tag]
        container = docker.run(tag, publish=[[8080, 8080]], detach=True)

        wait_for_truss("http://localhost:8080/v1/models/model", container)

        resp = requests.post(
            "http://localhost:8080/v1/models/model:predict",
            json=[0, 0, 0, 0, 0, 0, 0, 0, 0],
        )
        predictions = resp.json()["predictions"]
        assert len(predictions) == 1
        assert len(predictions[0]) == 1



import inspect
import types
from itertools import dropwhile
from typing import Optional, Set

import pkg_resources  # type: ignore
from pkg_resources import WorkingSet  # type: ignore

# Some packages are weird and have different
# imported names vs. system/pip names. Unfortunately,
# there is no systematic way to get pip names from
# a package's imported name. You'll have to add
# exceptions to this list manually!
POORLY_NAMED_PACKAGES = {"PIL": "Pillow", "sklearn": "scikit-learn"}

# We don't want a few foundation packages
IGNORED_PACKAGES = {"baseten", "pip", "truss", "pluggy", "pytest", "py"}

TOP_LEVEL_NAMESPACES_TO_DROP_FOR_INFERENCE = ["truss", "baseten"]


def infer_deps(must_include_deps: Optional[Set[str]] = None) -> Set[str]:
    """Infers the depedencies based on imports into the global namespace

    Args:
        must_include_deps (Set, optional):  The set of package names that
                                            must necessarily be imported.
                                            Defaults to None.

    Returns:
        Set[str]: set of required python requirements, including versions. E.g. `{"xgboost==1.6.1"}`
    """

    # Find the stack frame that likely has the relevant global inputs
    stack = inspect.stack()
    try:
        relevant_stack = _filter_truss_frames(stack)
    except StopIteration:
        return set()

    if not must_include_deps:
        must_include_deps = set()

    pkg_candidates = _extract_packages_from_frame(relevant_stack[0].frame)
    imports = must_include_deps.union(pkg_candidates)
    requirements = set([])

    # Must refresh working set manually to get latest installed
    pkg_resources.working_set = (
        WorkingSet._build_master()  # type: ignore # pylint: disable=protected-access
    )

    # Cross-check the names of installed packages vs. imported packages to get versions
    for pkg_in_frame in pkg_resources.working_set:
        if (
            pkg_in_frame.project_name in imports
            and pkg_in_frame.project_name not in IGNORED_PACKAGES
        ):
            requirements.add(f"{pkg_in_frame.project_name}=={pkg_in_frame.version}")
            # Remove the package from imports as it was added into requirements
            imports.remove(pkg_in_frame.project_name)

    # Add the must include deps not found in frame to requirements
    deps_not_found_in_frame = imports.intersection(must_include_deps)
    requirements = requirements.union(deps_not_found_in_frame)

    return requirements


def _filter_truss_frames(stack_frames):
    def is_truss_invocation_frame(stack_frame):
        module = inspect.getmodule(stack_frame.frame)
        if not module:
            return False

        module_name = module.__name__
        for namespace in TOP_LEVEL_NAMESPACES_TO_DROP_FOR_INFERENCE:
            if module_name.startswith(f"{namespace}."):
                return True
        return False

    return list(dropwhile(is_truss_invocation_frame, stack_frames))


def _extract_packages_from_frame(frame) -> Set[str]:
    candidate_symbols = {**frame.f_globals, **frame.f_locals}

    pkg_names = set()
    for name, val in candidate_symbols.items():
        if name.startswith("__"):
            continue

        if isinstance(val, types.ModuleType) and val.__name__ is not None:
            # Split ensures you get root package,
            # not just imported function
            pkg_name = val.__name__.split(".")[0]
        elif hasattr(val, "__module__") and val.__module__ is not None:
            pkg_name = val.__module__.split(".")[0]
        else:
            continue

        if pkg_name in POORLY_NAMED_PACKAGES:
            pkg_name = POORLY_NAMED_PACKAGES[pkg_name]

        pkg_names.add(pkg_name)

    return pkg_names



import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import pkg_resources
import yaml
from truss.constants import CONFIG_FILE
from truss.patch.hash import file_content_hash_str
from truss.patch.types import TrussSignature
from truss.templates.control.control.helpers.types import (
    Action,
    ModelCodePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_config import TrussConfig
from truss.truss_spec import TrussSpec

logger: logging.Logger = logging.getLogger(__name__)
PYCACHE_IGNORE_PATTERNS = [
    "**/__pycache__/**/*",
    "**/__pycache__/**",
]


def calc_truss_patch(
    truss_dir: Path,
    previous_truss_signature: TrussSignature,
    ignore_patterns: Optional[List[str]] = None,
) -> Optional[List[Patch]]:
    """
    Calculate patch for a truss from a previous state.

    Returns: None if patch cannot be calculated, otherwise a list of patches.
        Note that the none return value is pretty important, patch coverage is
        limited and this usually indicates that the identified change cannot be
        expressed with currently supported patches.

        Only standard and relevant truss paths are scanned for changes, rest
        ignored. E.g. at the root level, only changes to config.yaml are
        checked, any other changes are ignored.
    """

    def _relative_to_root(path: Path) -> str:
        return str(path.relative_to(truss_dir))

    if ignore_patterns is None:
        ignore_patterns = PYCACHE_IGNORE_PATTERNS

    changed_paths = _calc_changed_paths(
        truss_dir,
        previous_truss_signature.content_hashes_by_path,
        ignore_patterns,
    )
    # TODO(pankaj) Calculate model code patches only for now, add config changes
    # later.

    truss_spec = TrussSpec(truss_dir)
    model_module_path = _relative_to_root(truss_spec.model_module_dir)
    data_dir_path = _relative_to_root(truss_spec.data_dir)
    bundled_packages_path = _relative_to_root(truss_spec.bundled_packages_dir)

    def _under_unsupported_patch_dir(path: str) -> bool:
        """
        Checks if the given path is under one of the directories that don't
        support patching. Note that if path `is` one of those directories that's
        ok, because those empty directories can be ignored from patching point
        of view.
        """
        return _strictly_under(path, [data_dir_path, bundled_packages_path])

    patches = []
    for path in changed_paths["removed"]:
        if path.startswith(model_module_path):
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=Action.REMOVE,
                        path=_relative_to(path, model_module_path),
                    ),
                )
            )
        elif path == CONFIG_FILE:
            # Don't support removal of config file
            logger.info(f"Patching not supported for removing {path}")
            return None
        elif _under_unsupported_patch_dir(path):
            logger.info(f"Patching not supported for removing {path}")
            return None

    for path in changed_paths["added"] + changed_paths["updated"]:
        if path.startswith(model_module_path):
            full_path = truss_dir / path

            # TODO(pankaj) Add support for empty directories, skip them for now.
            if not full_path.is_file():
                continue

            action = Action.ADD if path in changed_paths["added"] else Action.UPDATE
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=action,
                        path=_relative_to(path, model_module_path),
                        content=_file_content(full_path),
                    ),
                )
            )
        elif path == CONFIG_FILE:
            new_config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
            prev_config = TrussConfig.from_dict(
                yaml.safe_load(previous_truss_signature.config)
            )
            config_patches = _calc_config_patches(prev_config, new_config)
            if config_patches is None:
                logger.info(f"Unable to patch update to {path}")
                return None
            patches.extend(config_patches)
        elif _under_unsupported_patch_dir(path):
            logger.info(f"Patching not supported for updating {path}")
            return None
    return patches


def _calc_changed_paths(
    root: Path,
    previous_root_path_content_hashes: Dict[str, str],
    ignore_patterns: Optional[List[str]],
) -> Dict[str, List[str]]:
    """
    TODO(pankaj) add support for directory creation in patch
    """
    root_relative_new_paths = set(
        (str(path.relative_to(root)) for path in root.glob("**/*"))
    )
    unignored_new_paths = _calc_unignored_paths(
        root, root_relative_new_paths, ignore_patterns
    )
    previous_root_relative_paths = set(previous_root_path_content_hashes.keys())
    unignored_prev_paths = _calc_unignored_paths(
        root, previous_root_relative_paths, ignore_patterns
    )

    added_paths = unignored_new_paths - unignored_prev_paths
    removed_paths = unignored_prev_paths - unignored_new_paths

    updated_paths = set()
    common_paths = unignored_new_paths.intersection(unignored_prev_paths)
    for path in common_paths:
        full_path: Path = root / path
        if full_path.is_file():
            content_hash = file_content_hash_str(full_path)
            previous_content_hash = previous_root_path_content_hashes[path]
            if content_hash != previous_content_hash:
                updated_paths.add(path)

    return {
        "added": list(added_paths),
        "updated": list(updated_paths),
        "removed": list(removed_paths),
    }


def _calc_unignored_paths(
    root: Path,
    root_relative_paths: Set[str],
    ignore_patterns: Optional[List[str]] = None,
) -> Set[str]:
    root_relative_ignored_paths = set()
    if ignore_patterns is not None:
        for ignore_pattern in ignore_patterns:
            ignored_paths_for_pattern = set(
                (str(path.relative_to(root)) for path in root.glob(ignore_pattern))
            )
            root_relative_ignored_paths.update(ignored_paths_for_pattern)

    return root_relative_paths - root_relative_ignored_paths


def _calc_config_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> Optional[List[Patch]]:
    """Calculate patch based on changes to config.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    # Currently only calculate patches for python requirements and system
    # packages, bail out if anything else has changed.
    if not _only_expected_config_differences(prev_config, new_config):
        return None

    python_requirement_patches = _calc_python_requirements_patches(
        prev_config, new_config
    )
    system_package_patches = _calc_system_packages_patches(prev_config, new_config)
    return [*python_requirement_patches, *system_package_patches]


def _calc_python_requirements_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to python requirements.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    patches = []
    prev_reqs = _parsed_reqs_by_name(prev_config.requirements)
    prev_req_names = set(prev_reqs.keys())
    new_reqs = _parsed_reqs_by_name(new_config.requirements)
    new_req_names = set(new_reqs.keys())
    removed_reqs = prev_req_names.difference(new_req_names)
    for removed_req in removed_reqs:
        patches.append(_mk_python_requirement_patch(Action.REMOVE, removed_req))

    added_reqs = new_req_names.difference(prev_req_names)
    for added_req in added_reqs:
        patches.append(
            _mk_python_requirement_patch(Action.ADD, str(new_reqs[added_req]))
        )

    for req in new_req_names.intersection(prev_req_names):
        if new_reqs[req] != prev_reqs[req]:
            patches.append(
                _mk_python_requirement_patch(Action.UPDATE, str(new_reqs[req]))
            )

    return patches


def _calc_system_packages_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to system packates.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    patches = []
    prev_pkgs = _system_pacakges_set(prev_config)
    new_pkgs = _system_pacakges_set(new_config)
    removed_pkgs = prev_pkgs.difference(new_pkgs)
    for removed_pkg in removed_pkgs:
        patches.append(_mk_system_package_patch(Action.REMOVE, removed_pkg))

    added_pkgs = new_pkgs.difference(prev_pkgs)
    for added_pkg in added_pkgs:
        patches.append(_mk_system_package_patch(Action.ADD, added_pkg))

    return patches


def _system_pacakges_set(config: TrussConfig) -> Set[str]:
    pkgs = []
    for sys_pkg_line in config.system_packages:
        pkgs.extend(sys_pkg_line.strip().split())
    return set(pkgs)


def _parsed_reqs_by_name(reqs: List[str]) -> Dict[str, Any]:
    parsed_reqs_by_name = {}
    for req in reqs:
        parsed_req = pkg_resources.Requirement.parse(req)
        parsed_reqs_by_name[parsed_req.name] = parsed_req  # type: ignore
    return parsed_reqs_by_name


def _only_expected_config_differences(
    prev_config: TrussConfig, new_config: TrussConfig
) -> bool:
    prev_config_dict = prev_config.to_dict()
    prev_config_dict["requirements"] = []
    prev_config_dict["system_packages"] = []

    new_config_dict = new_config.to_dict()
    new_config_dict["requirements"] = []
    new_config_dict["system_packages"] = []

    return prev_config_dict == new_config_dict


def _mk_python_requirement_patch(action: Action, requirement: str) -> Patch:
    return Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=action,
            requirement=requirement,
        ),
    )


def _mk_system_package_patch(action: Action, package: str) -> Patch:
    return Patch(
        type=PatchType.SYSTEM_PACKAGE,
        body=SystemPackagePatch(
            action=action,
            package=package,
        ),
    )


def _relative_to(path: str, relative_to_path: str):
    return str(Path(path).relative_to(relative_to_path))


def _strictly_under(path: str, parent_paths: List[str]) -> bool:
    """
    Checks if given path is under one of the given paths, but not the same as
    them. Assumes that parent paths themselves are not under each other.
    """
    for dir_path in parent_paths:
        if path.startswith(dir_path) and not path == dir_path:
            return True
    return False


def _file_content(path: Path) -> str:
    with path.open() as file:
        return file.read()



from typing import Any

import mlflow
import numpy as np


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        config = kwargs["config"]
        model_metadata = config["model_metadata"]
        self._model_binary_dir = model_metadata["model_binary_dir"]
        self._model = None

    def load(self):
        model_binary_dir_path = self._data_dir / self._model_binary_dir
        self._model = mlflow.pyfunc.load_model(model_binary_dir_path / "model")

    def predict(self, model_input: Any) -> Any:
        model_output = {}
        inputs = np.array(model_input)
        result = self._model.predict(inputs).tolist()
        model_output["predictions"] = result
        return model_output



from truss.build import from_directory, mk_truss


def test_mk_truss_passthrough(sklearn_rfc_model, tmp_path):
    dir_path = tmp_path / "truss"
    data_file_path = tmp_path / "data.txt"
    with data_file_path.open("w") as data_file:
        data_file.write("test")
    req_file_path = tmp_path / "requirements.txt"
    requirements = [
        "tensorflow==2.3.1",
        "uvicorn==0.12.2",
    ]
    with req_file_path.open("w") as req_file:
        for req in requirements:
            req_file.write(f"{req}\n")
    scaf = mk_truss(
        sklearn_rfc_model,
        target_directory=dir_path,
        data_files=[str(data_file_path)],
        requirements_file=str(req_file_path),
    )
    spec = scaf.spec
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert (spec.data_dir / "data.txt").exists()
    assert spec.requirements == requirements


def test_from_directory_passthrough(sklearn_rfc_model, tmp_path):
    dir_path = tmp_path / "truss"
    data_file_path = tmp_path / "data.txt"
    with data_file_path.open("w") as data_file:
        data_file.write("test")
    req_file_path = tmp_path / "requirements.txt"
    requirements = [
        "tensorflow==2.3.1",
        "uvicorn==0.12.2",
    ]
    with req_file_path.open("w") as req_file:
        for req in requirements:
            req_file.write(f"{req}\n")
    mk_truss(
        sklearn_rfc_model,
        target_directory=dir_path,
        data_files=[str(data_file_path)],
        requirements_file=str(req_file_path),
    )
    scaf = from_directory(dir_path)
    spec = scaf.spec
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert (spec.data_dir / "data.txt").exists()
    assert spec.requirements == requirements



from pathlib import Path
from typing import Dict

from truss.patch.hash import file_content_hash_str


def directory_content_signature(root: Path) -> Dict[str, str]:
    """Calculate content signature of a filesystem directory.

    Sort all files by path, store file path with content hash.
    Signatures are meant to track changes in directory. e.g.
    A previous signature and the directory can be combined to create
    a patch from the previous state.

    Hash of directories is marked None.
    """
    paths = list(root.glob("**/*"))
    paths.sort(key=lambda p: p.relative_to(root))

    def path_hash(pth: Path):
        if pth.is_file():
            return file_content_hash_str(pth)

    return {str(path.relative_to(root)): path_hash(path) for path in paths}



import os
from typing import Dict

import yaml
from common.logging import setup_logging
from common.truss_server import TrussServer  # noqa: E402

CONFIG_FILE = "config.yaml"

setup_logging()


class ConfiguredTrussServer:
    _config: Dict
    _port: int

    def __init__(self, config_path: str, port: int):
        self._port = port
        with open(config_path, encoding="utf-8") as config_file:
            self._config = yaml.safe_load(config_file)

    def start(self):
        server = TrussServer(http_port=self._port, config=self._config)
        server.start_model()


if __name__ == "__main__":
    env_port = int(os.environ.get("INFERENCE_SERVER_PORT", "8080"))
    ConfiguredTrussServer(CONFIG_FILE, env_port).start()



from typing import Any

from truss.errors import FrameworkNotSupportedError
from truss.model_framework import ModelFramework
from truss.model_frameworks.huggingface_transformer import HuggingfaceTransformer
from truss.model_frameworks.keras import Keras
from truss.model_frameworks.lightgbm import LightGBM
from truss.model_frameworks.mlflow import Mlflow
from truss.model_frameworks.pytorch import PyTorch
from truss.model_frameworks.sklearn import SKLearn
from truss.model_frameworks.xgboost import XGBoost
from truss.types import ModelFrameworkType

MODEL_FRAMEWORKS_BY_TYPE = {
    ModelFrameworkType.SKLEARN: SKLearn(),
    ModelFrameworkType.KERAS: Keras(),
    ModelFrameworkType.HUGGINGFACE_TRANSFORMER: HuggingfaceTransformer(),
    ModelFrameworkType.PYTORCH: PyTorch(),
    ModelFrameworkType.XGBOOST: XGBoost(),
    ModelFrameworkType.LIGHTGBM: LightGBM(),
    ModelFrameworkType.MLFLOW: Mlflow(),
}


SUPPORTED_MODEL_FRAMEWORKS = [
    ModelFrameworkType.SKLEARN,
    ModelFrameworkType.KERAS,
    ModelFrameworkType.TENSORFLOW,
    ModelFrameworkType.HUGGINGFACE_TRANSFORMER,
    ModelFrameworkType.XGBOOST,
    ModelFrameworkType.PYTORCH,
    ModelFrameworkType.LIGHTGBM,
]


def model_framework_from_model(model: Any) -> ModelFramework:
    return model_framework_from_model_class(model.__class__)


def model_framework_from_model_class(model_class) -> ModelFramework:
    for model_framework in MODEL_FRAMEWORKS_BY_TYPE.values():
        if model_framework.supports_model_class(model_class):
            return model_framework

    raise FrameworkNotSupportedError(
        "Model must be one of "
        + "/".join([t.value for t in SUPPORTED_MODEL_FRAMEWORKS])
    )



import logging
import tempfile
from pathlib import Path

import numpy as np
import pytest
from truss.constants import PYTORCH
from truss.model_frameworks import SKLearn
from truss.model_inference import (
    infer_model_information,
    map_to_supported_python_version,
    validate_provided_parameters_with_model,
)
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_handle import TrussHandle

logger = logging.getLogger(__name__)


def test_pytorch_init_arg_validation(
    pytorch_model_with_init_args, pytorch_model_init_args
):
    pytorch_model_with_init_args, _ = pytorch_model_with_init_args
    # Validates with args and kwargs
    validate_provided_parameters_with_model(
        pytorch_model_with_init_args.__class__, pytorch_model_init_args
    )

    # Errors if bad args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(
            pytorch_model_with_init_args.__class__, {"foo": "bar"}
        )

    # Validates with only args
    copied_args = pytorch_model_init_args.copy()
    copied_args.pop("kwarg1")
    copied_args.pop("kwarg2")
    validate_provided_parameters_with_model(pytorch_model_with_init_args, copied_args)

    # Requires all args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(pytorch_model_with_init_args, {})


def test_infer_model_information(pytorch_model_with_init_args):
    model_info = infer_model_information(pytorch_model_with_init_args[0])
    assert model_info.model_framework == PYTORCH
    assert model_info.model_type == "MyModel"


@pytest.mark.parametrize(
    "python_version, expected_python_version",
    [
        ("py37", "py37"),
        ("py38", "py38"),
        ("py39", "py39"),
        ("py310", "py39"),
        ("py311", "py39"),
        ("py36", "py37"),
    ],
)
def test_map_to_supported_python_version(python_version, expected_python_version):
    out_python_version = map_to_supported_python_version(python_version)
    assert out_python_version == expected_python_version


@pytest.mark.integration
def test_binary_request(sklearn_rfc_model):
    with ensure_kill_all(), tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        sklearn_framework = SKLearn()
        sklearn_framework.to_truss(sklearn_rfc_model, truss_dir)
        tr = TrussHandle(truss_dir)
        predictions = tr.docker_predict([[0, 0, 0, 0]], local_port=8090, binary=True)
        assert len(predictions["probabilities"]) == 1
        assert np.shape(predictions["probabilities"]) == (1, 3)



from pathlib import Path
from typing import Dict, Set

from truss.constants import TENSORFLOW_REQ_MODULE_NAMES
from truss.model_framework import ModelFramework
from truss.types import ModelFrameworkType


class Keras(ModelFramework):
    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.KERAS

    def required_python_depedencies(self) -> Set[str]:
        return TENSORFLOW_REQ_MODULE_NAMES

    def serialize_model_to_directory(self, model, target_directory: Path):
        model.save(target_directory)

    def model_metadata(self, model) -> Dict[str, str]:
        return {
            "model_binary_dir": "model",
        }

    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition(".")
        return model_framework in [
            ModelFrameworkType.KERAS.value,
            ModelFrameworkType.TENSORFLOW.value,
        ]



from unittest.mock import patch

import pytest
from truss.templates.server.common.external_data_resolver import (
    _external_data_items,
    download_external_data,
)
from truss.truss_config import ExternalData, ExternalDataItem, TrussConfig


def test_external_data_items(tmp_path):
    url1 = "http://dummy1"
    url2 = "http://dummy2"
    data_dir = tmp_path
    config = TrussConfig(
        external_data=ExternalData(
            items=[
                ExternalDataItem(url=url1, local_data_path="dummy1"),
                ExternalDataItem(url=url2, local_data_path="dir/dummy2"),
            ]
        )
    )

    items = _external_data_items(data_dir, config.to_dict())
    assert items == [
        (url1, tmp_path / "dummy1"),
        (url2, tmp_path / "dir" / "dummy2"),
    ]


def test_external_data_items_unsupported_backend(tmp_path):
    url = "http://dummy1"
    data_dir = tmp_path
    config = TrussConfig(
        external_data=ExternalData(
            items=[
                ExternalDataItem(
                    url=url, local_data_path="dummy1", backend="unsupported"
                ),
            ]
        )
    )

    with pytest.raises(ValueError):
        _external_data_items(data_dir, config.to_dict())


def test_external_data_items_empty(tmp_path):
    data_dir = tmp_path
    config = TrussConfig()

    items = _external_data_items(data_dir, config.to_dict())
    assert items == []


@patch("truss.templates.server.common.external_data_resolver._try_b10cp")
@patch("truss.templates.server.common.external_data_resolver._download_using_requests")
def test_download_external_data(
    _try_b10cp_mock, _download_using_requests_mock, tmp_path
):
    _try_b10cp_mock.return_value = False
    url = "http://dummy1"
    data_dir = tmp_path
    config = TrussConfig(
        external_data=ExternalData(
            items=[
                ExternalDataItem(url=url, local_data_path="dummy1"),
            ]
        )
    )
    download_external_data(data_dir, config.to_dict())
    assert _try_b10cp_mock.called_with(url, tmp_path / "dummy1")
    assert _download_using_requests_mock.called_with(url, tmp_path / "dummy1")



import pytest
from truss.truss_config import (
    DEFAULT_CPU,
    DEFAULT_MEMORY,
    DEFAULT_USE_GPU,
    Accelerator,
    AcceleratorSpec,
    Resources,
)


@pytest.mark.parametrize(
    "input_dict, expect_resources, output_dict",
    [
        (
            {},
            Resources(),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": DEFAULT_USE_GPU,
                "accelerator": None,
            },
        ),
        (
            {"accelerator": None},
            Resources(),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": DEFAULT_USE_GPU,
                "accelerator": None,
            },
        ),
        (
            {"accelerator": "V100"},
            Resources(accelerator=AcceleratorSpec(Accelerator.V100, 1), use_gpu=True),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": True,
                "accelerator": "V100",
            },
        ),
        (
            {"accelerator": "T4:1"},
            Resources(accelerator=AcceleratorSpec(Accelerator.T4, 1), use_gpu=True),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": True,
                "accelerator": "T4",
            },
        ),
        (
            {"accelerator": "A10G:4"},
            Resources(accelerator=AcceleratorSpec(Accelerator.A10G, 4), use_gpu=True),
            {
                "cpu": DEFAULT_CPU,
                "memory": DEFAULT_MEMORY,
                "use_gpu": True,
                "accelerator": "A10G:4",
            },
        ),
    ],
)
def test_parse_resources(input_dict, expect_resources, output_dict):
    parsed_result = Resources.from_dict(input_dict)
    assert parsed_result == expect_resources
    assert parsed_result.to_dict() == output_dict


@pytest.mark.parametrize(
    "input_str, expected_acc",
    [
        (None, AcceleratorSpec(None, 0)),
        ("T4", AcceleratorSpec(Accelerator.T4, 1)),
        ("A10G:4", AcceleratorSpec(Accelerator.A10G, 4)),
        ("A100:8", AcceleratorSpec(Accelerator.A100, 8)),
    ],
)
def test_acc_spec_from_str(input_str, expected_acc):
    assert AcceleratorSpec.from_str(input_str) == expected_acc
